@@ -8,9 +8,9 @@ use Lemonldap::NG::Portal::Main::Constants qw(
   PE_MISSINGREQATTR
 );
 
-extends 'Lemonldap::NG::Common::Module';
+extends 'Lemonldap::NG::Portal::Main::UserDB';
 
-our $VERSION = '2.0.12';
+our $VERSION = '2.19.0';
 
 # INITIALIZATION
 
@@ -56,8 +56,4 @@ sub setSessionInfo {
     return PE_OK;
 }
 
-sub setGroups {
-    return PE_OK;
-}
-
 1;
